# coding: utf-8

"""
    PKS

    PKS API  # noqa: E501

    OpenAPI spec version: 1.1.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class KubernetesProfile(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'uuid': 'str',
        'owner': 'str',
        'description': 'str',
        'created_at': 'int',
        'customizations': 'list[KubernetesComponentCustomization]',
        'experimental_customizations': 'list[KubernetesComponentCustomization]'
    }

    attribute_map = {
        'name': 'name',
        'uuid': 'uuid',
        'owner': 'owner',
        'description': 'description',
        'created_at': 'created_at',
        'customizations': 'customizations',
        'experimental_customizations': 'experimental_customizations'
    }

    def __init__(self, name=None, uuid=None, owner=None, description=None, created_at=None, customizations=None, experimental_customizations=None):  # noqa: E501
        """KubernetesProfile - a model defined in Swagger"""  # noqa: E501

        self._name = None
        self._uuid = None
        self._owner = None
        self._description = None
        self._created_at = None
        self._customizations = None
        self._experimental_customizations = None
        self.discriminator = None

        self.name = name
        self.uuid = uuid
        if owner is not None:
            self.owner = owner
        if description is not None:
            self.description = description
        if created_at is not None:
            self.created_at = created_at
        if customizations is not None:
            self.customizations = customizations
        if experimental_customizations is not None:
            self.experimental_customizations = experimental_customizations

    @property
    def name(self):
        """Gets the name of this KubernetesProfile.  # noqa: E501


        :return: The name of this KubernetesProfile.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this KubernetesProfile.


        :param name: The name of this KubernetesProfile.  # noqa: E501
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def uuid(self):
        """Gets the uuid of this KubernetesProfile.  # noqa: E501


        :return: The uuid of this KubernetesProfile.  # noqa: E501
        :rtype: str
        """
        return self._uuid

    @uuid.setter
    def uuid(self, uuid):
        """Sets the uuid of this KubernetesProfile.


        :param uuid: The uuid of this KubernetesProfile.  # noqa: E501
        :type: str
        """
        if uuid is None:
            raise ValueError("Invalid value for `uuid`, must not be `None`")  # noqa: E501

        self._uuid = uuid

    @property
    def owner(self):
        """Gets the owner of this KubernetesProfile.  # noqa: E501


        :return: The owner of this KubernetesProfile.  # noqa: E501
        :rtype: str
        """
        return self._owner

    @owner.setter
    def owner(self, owner):
        """Sets the owner of this KubernetesProfile.


        :param owner: The owner of this KubernetesProfile.  # noqa: E501
        :type: str
        """

        self._owner = owner

    @property
    def description(self):
        """Gets the description of this KubernetesProfile.  # noqa: E501


        :return: The description of this KubernetesProfile.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this KubernetesProfile.


        :param description: The description of this KubernetesProfile.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def created_at(self):
        """Gets the created_at of this KubernetesProfile.  # noqa: E501


        :return: The created_at of this KubernetesProfile.  # noqa: E501
        :rtype: int
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this KubernetesProfile.


        :param created_at: The created_at of this KubernetesProfile.  # noqa: E501
        :type: int
        """

        self._created_at = created_at

    @property
    def customizations(self):
        """Gets the customizations of this KubernetesProfile.  # noqa: E501


        :return: The customizations of this KubernetesProfile.  # noqa: E501
        :rtype: list[KubernetesComponentCustomization]
        """
        return self._customizations

    @customizations.setter
    def customizations(self, customizations):
        """Sets the customizations of this KubernetesProfile.


        :param customizations: The customizations of this KubernetesProfile.  # noqa: E501
        :type: list[KubernetesComponentCustomization]
        """

        self._customizations = customizations

    @property
    def experimental_customizations(self):
        """Gets the experimental_customizations of this KubernetesProfile.  # noqa: E501


        :return: The experimental_customizations of this KubernetesProfile.  # noqa: E501
        :rtype: list[KubernetesComponentCustomization]
        """
        return self._experimental_customizations

    @experimental_customizations.setter
    def experimental_customizations(self, experimental_customizations):
        """Sets the experimental_customizations of this KubernetesProfile.


        :param experimental_customizations: The experimental_customizations of this KubernetesProfile.  # noqa: E501
        :type: list[KubernetesComponentCustomization]
        """

        self._experimental_customizations = experimental_customizations

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(KubernetesProfile, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, KubernetesProfile):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
